import warnings

warnings.warn(
    "`cubed.extensions.mem_warn` is deprecated, please use `cubed.diagnostics.mem_warn` instead",
    DeprecationWarning,
)

from cubed.diagnostics.mem_warn import MemoryWarningCallback  # noqa: F401



import warnings

warnings.warn(
    "`cubed.extensions.rich` is deprecated, please use `cubed.diagnostics.rich` instead",
    DeprecationWarning,
)

from cubed.diagnostics.rich import RichProgressBar  # noqa: F401



from numpy.testing import assert_array_equal

import cubed.array_api as xp


# This is less strict than the spec, but is supported by implementations like NumPy
def test_prod_sum_bool():
    a = xp.ones((2,), dtype=xp.bool)
    assert_array_equal(xp.prod(a).compute(), xp.asarray([1], dtype=xp.int64))
    assert_array_equal(xp.sum(a).compute(), xp.asarray([2], dtype=xp.int64))



from typing import Optional

from cubed import config
from cubed.types import T_DType, T_RegularChunks, T_Shape, T_Store


def open_backend_array(
    store: T_Store,
    mode: str,
    *,
    shape: Optional[T_Shape] = None,
    dtype: Optional[T_DType] = None,
    chunks: Optional[T_RegularChunks] = None,
    path: Optional[str] = None,
    **kwargs,
):
    # get storage name from top-level config
    # e.g. set globally with CUBED_STORAGE_NAME=tensorstore
    storage_name = config.get("storage_name", None)

    if storage_name is None:
        import zarr

        if zarr.__version__[0] == "3":
            storage_name = "zarr-python-v3"
        else:
            storage_name = "zarr-python"

    if storage_name == "zarr-python":
        from cubed.storage.backends.zarr_python import open_zarr_array

        open_func = open_zarr_array
    elif storage_name == "zarr-python-v3":
        from cubed.storage.backends.zarr_python_v3 import open_zarr_v3_array

        open_func = open_zarr_v3_array
    elif storage_name == "tensorstore":
        from cubed.storage.backends.tensorstore import open_tensorstore_array

        open_func = open_tensorstore_array
    else:
        raise ValueError(f"Unrecognized storage name: {storage_name}")

    return open_func(
        store,
        mode,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        path=path,
        **kwargs,
    )



import time
from contextlib import nullcontext
from functools import partial
from itertools import islice
from pathlib import Path

from cubed.runtime.types import OperationStartEvent, TaskEndEvent
from cubed.utils import peak_measured_mem

try:
    import memray
except ImportError:
    memray = None  # type: ignore

sym_counter = 0


def gensym(name: str) -> str:
    global sym_counter
    sym_counter += 1
    return f"{name}-{sym_counter:03}"


def execute_with_stats(function, *args, **kwargs):
    """Invoke function and measure timing information and peak memory usage.

    Returns the result of the function call and a stats dictionary.
    """

    peak_measured_mem_start = peak_measured_mem()
    function_start_tstamp = time.time()
    result = function(*args, **kwargs)
    function_end_tstamp = time.time()
    peak_measured_mem_end = peak_measured_mem()
    return result, dict(
        function_start_tstamp=function_start_tstamp,
        function_end_tstamp=function_end_tstamp,
        peak_measured_mem_start=peak_measured_mem_start,
        peak_measured_mem_end=peak_measured_mem_end,
    )


def execution_stats(func):
    """Decorator to measure timing information and peak memory usage of a function call."""

    return partial(execute_with_stats, func)


def execute_with_memray(function, input, **kwargs):
    # only run memray if installed, and only for first input (for operations that run on block locations)
    if (
        memray is not None
        and "compute_id" in kwargs
        and isinstance(input, list)
        and all(isinstance(i, int) for i in input)
        and sum(input) == 0
    ):
        compute_id = kwargs["compute_id"]
        name = kwargs["name"]
        memray_dir = Path(f"history/{compute_id}/memray")
        memray_dir.mkdir(parents=True, exist_ok=True)
        cm = memray.Tracker(memray_dir / f"{name}.bin")
    else:
        cm = nullcontext()
    with cm:
        result = result = function(input, **kwargs)
        return result


def profile_memray(func):
    """Decorator to profile a function call with memray."""
    return partial(execute_with_memray, func)


def handle_operation_start_callbacks(callbacks, name):
    if callbacks is not None:
        event = OperationStartEvent(name)
        [callback.on_operation_start(event) for callback in callbacks]


def handle_callbacks(callbacks, stats):
    """Construct a TaskEndEvent from stats and send to all callbacks."""

    if callbacks is not None:
        if "task_result_tstamp" not in stats:
            task_result_tstamp = time.time()
            event = TaskEndEvent(
                task_result_tstamp=task_result_tstamp,
                **stats,
            )
        else:
            event = TaskEndEvent(**stats)
        [callback.on_task_end(event) for callback in callbacks]


# this will be in Python 3.12 https://docs.python.org/3.12/library/itertools.html#itertools.batched
def batched(iterable, n):
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch



from typing import Optional, Union

import zarr
from numcodecs.registry import get_codec

from cubed.types import T_DType, T_RegularChunks, T_Shape, T_Store


def open_zarr_array(
    store: T_Store,
    mode: str,
    *,
    shape: Optional[T_Shape] = None,
    dtype: Optional[T_DType] = None,
    chunks: Optional[T_RegularChunks] = None,
    path: Optional[str] = None,
    compressor: Union[dict, str, None] = "default",
    **kwargs,
):
    if isinstance(compressor, dict):
        compressor = get_codec(compressor)

    return zarr.open_array(
        store,
        mode=mode,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        path=path,
        compressor=compressor,
        **kwargs,
    )



import warnings

warnings.warn(
    "`cubed.extensions.history` is deprecated, please use `cubed.diagnostics.history` instead",
    DeprecationWarning,
)

from cubed.diagnostics.history import HistoryCallback  # noqa: F401



from typing import Optional

import zarr

from cubed.types import T_DType, T_RegularChunks, T_Shape, T_Store


class ZarrV3ArrayGroup(dict):
    def __init__(
        self,
        shape: Optional[T_Shape] = None,
        dtype: Optional[T_DType] = None,
        chunks: Optional[T_RegularChunks] = None,
    ):
        dict.__init__(self)
        self.shape = shape
        self.dtype = dtype
        self.chunks = chunks

    def __getitem__(self, key):
        if isinstance(key, str):
            return super().__getitem__(key)
        return {field: zarray[key] for field, zarray in self.items()}

    def set_basic_selection(self, selection, value, fields=None):
        self[fields][selection] = value


def open_zarr_v3_array(
    store: T_Store,
    mode: str,
    *,
    shape: Optional[T_Shape] = None,
    dtype: Optional[T_DType] = None,
    chunks: Optional[T_RegularChunks] = None,
    path: Optional[str] = None,
    **kwargs,
):
    if isinstance(chunks, int):
        chunks = (chunks,)

    if dtype is None or not hasattr(dtype, "fields") or dtype.fields is None:
        return zarr.open(
            store=store,
            mode=mode,
            shape=shape,
            dtype=dtype,
            chunks=chunks,
            path=path,
        )

    group = zarr.open_group(store=store, mode=mode, path=path)

    # create/open all the arrays in the group
    ret = ZarrV3ArrayGroup(shape=shape, dtype=dtype, chunks=chunks)
    for field in dtype.fields:
        field_dtype, _ = dtype.fields[field]
        if mode in ("r", "r+"):
            ret[field] = group[field]
        else:
            ret[field] = group.create_array(
                field,
                shape=shape,
                dtype=field_dtype,
                chunk_shape=chunks,
            )
    return ret



import warnings

warnings.warn(
    "`cubed.extensions.tqdm` is deprecated, please use `cubed.diagnostics.tqdm` instead",
    DeprecationWarning,
)

from cubed.diagnostics.tqdm import (  # noqa: F401
    TqdmProgressBar,
    std_out_err_redirect_tqdm,
)



import time
from dataclasses import asdict
from pathlib import Path
from typing import Optional

import matplotlib.patches as mpatches
import numpy as np
import pandas as pd
import pylab
import seaborn as sns

from cubed.runtime.types import Callback

sns.set_style("whitegrid")
pylab.switch_backend("Agg")


class TimelineVisualizationCallback(Callback):
    def __init__(self, format: Optional[str] = "svg") -> None:
        self.format = format

    def on_compute_start(self, event):
        self.start_tstamp = time.time()
        self.stats = []

    def on_task_end(self, event):
        self.stats.append(asdict(event))

    def on_compute_end(self, event):
        self.end_tstamp = time.time()

        stats_df = pd.DataFrame(self.stats)
        stats_df = stats_df.sort_values(
            by=["task_create_tstamp", "name"], ascending=True
        )
        total_calls = len(stats_df)
        palette = sns.color_palette("deep", 6)

        fig = pylab.figure(figsize=(10, 6))
        ax = fig.add_subplot(1, 1, 1)

        y = np.arange(total_calls)
        point_size = 10

        fields = [
            ("task create", stats_df.task_create_tstamp - self.start_tstamp),
            ("function start", stats_df.function_start_tstamp - self.start_tstamp),
            ("function end", stats_df.function_end_tstamp - self.start_tstamp),
            ("task result", stats_df.task_result_tstamp - self.start_tstamp),
        ]

        patches = []
        for f_i, (field_name, val) in enumerate(fields):
            ax.scatter(
                val, y, c=[palette[f_i]], edgecolor="none", s=point_size, alpha=0.8
            )
            patches.append(mpatches.Patch(color=palette[f_i], label=field_name))

        ax.set_xlabel("Execution Time (sec)")
        ax.set_ylabel("Function Call")

        legend = pylab.legend(handles=patches, loc="upper right", frameon=True)
        legend.get_frame().set_facecolor("#FFFFFF")

        yplot_step = int(np.max([1, total_calls / 20]))
        y_ticks = np.arange(total_calls // yplot_step + 2) * yplot_step
        ax.set_yticks(y_ticks)
        ax.set_ylim(-0.02 * total_calls, total_calls * 1.02)
        for y in y_ticks:
            ax.axhline(y, c="k", alpha=0.1, linewidth=1)

        max_seconds = np.max(self.end_tstamp - self.start_tstamp) * 1.25
        xplot_step = max(int(max_seconds / 8), 1)
        x_ticks = np.arange(max_seconds // xplot_step + 2) * xplot_step
        ax.set_xlim(0, max_seconds)

        ax.set_xticks(x_ticks)
        for x in x_ticks:
            ax.axvline(x, c="k", alpha=0.2, linewidth=0.8)

        ax.grid(False)
        fig.tight_layout()

        self.dst = Path(f"history/{event.compute_id}")
        self.dst.mkdir(parents=True, exist_ok=True)
        self.dst = self.dst / f"timeline.{self.format}"

        fig.savefig(self.dst)



import warnings

warnings.warn(
    "`cubed.extensions.timeline` is deprecated, please use `cubed.diagnostics.timeline` instead",
    DeprecationWarning,
)

from cubed.diagnostics.timeline import TimelineVisualizationCallback  # noqa: F401
